/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Son Okt 20 11:24:44 CEST 2002
    copyright            : (C) 2002 by
    email                :
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <sstream>
#include <stdlib.h>
#include <sys/wait.h>
#include "noad.h"

const char *recDir = NULL;
bool bFork = false;

#define noMARK_CHECK
void doNoad(bool isLive, const char *fname)
{
  time_t              start;
  time_t              end;
  int iNumFrames = -1;
  int iNewNumFrames = -1;
  #ifdef MARK_CHECK
    start = time(NULL);
    syslog(LOG_INFO, "%s start noad%s for %s ", myTime(start), getVersion(), fname);
    fprintf(stderr,"%s start noad%s for %s\n", myTime(start), getVersion(), fname);
    noadData *pdata = new noadData();
    pdata->initBuffer();
    checkMarksOnIFrames(pdata, fname);
    end = time(NULL);
    fprintf(stderr,"%s noad done for %s (%ld secs)\n", myTime(end), fname,end-start);
    syslog(LOG_INFO, "%s noad done for %s (%ld secs)", myTime(end), fname,end-start);
    delete pdata;
  return;
  #endif
  do
  {
    if( isLive )
      sleep(300);
    noadData *pdata = new noadData();
    pdata->initBuffer();
    start = time(NULL);
    syslog(LOG_INFO, "%s start noad%s for %s ", myTime(start), getVersion(), fname);
    fprintf(stderr,"%s start noad%s for %s\n", myTime(start), getVersion(), fname);
    iNumFrames = iNewNumFrames;
    iNewNumFrames = doX11Scan(pdata, fname, iNumFrames);
    if( cctrl != NULL )
      delete cctrl;
    cctrl = NULL;
    delete pdata;
    end = time(NULL);
    fprintf(stderr,"%s noad done for %s (%s)\n", myTime(end), fname,secToTime(end-start));
    syslog(LOG_INFO, "%s noad done for %s (%s)", myTime(end), fname,secToTime(end-start));
  }while(isLive == true && iNewNumFrames > iNumFrames );
}

void writeStatistic(const char *statfilename, const char *recording)
{
  FILE *f = fopen(statfilename, "a+");
  if (!f)
  {
    LOG_ERROR_STR(statfilename);
    return;
  }
  time_t start = time(NULL);
  fprintf(f,"%s;%s;%s,%s,%s;%dx%d;%d (%s);%d;%d;%d;%d;%d;%d;%s;%s\n",
//  fprintf(f,"%s;%s;%s;%dx%d;%d (%s);%d;%d;%d;%d;%d;%d;%s;%s\n",
      myTime(start),
      getVersion(),
      hasBlackLines?"yes":"no",
      hasAC3?"yes":"no",
      hasOverlaps?"yes":"no",
      picWidth,picHeight,
      totalFrames,
      IndexToHMSF(totalFrames,false),
      totalDecodedFrames,
      decodedFramesForLogoDetection,
      secsForLogoDetection,
      decodedFramesForLogoCheck,
      secsForLogoCheck,
      secsForScan,
      recording,
      "add your comment here");
  fclose(f);
}

int fork2()
{
    pid_t pid;
    //int rc;
    int status;

    if (!(pid = fork()))
    {
        switch (fork())
        {
          case 0:  return 0;
          case -1: _exit(errno);    /* assumes all errnos are <256 */
          default: _exit(0);
        }
    }

    if (pid < 0 || waitpid(pid,&status,0) < 0)
      return -1;

    if (WIFEXITED(status))
      if (WEXITSTATUS(status) == 0)
        return 1;
      else
        errno = WEXITSTATUS(status);
    else
      errno = EINTR;  /* well, sort of :-) */

    return -1;
}

static int close_files(void)
{
  struct rlimit file_limit;
  int file_max;

  if (getrlimit(RLIMIT_NOFILE, &file_limit) == -1)
  {
    perror("getrlimit");
    return !0;
  }
  file_max = (int)file_limit.rlim_cur;

  for (int f = 0; f < file_max; f++)
  {
    if (f != fileno(stdin) && f != fileno(stdout) && f != fileno(stderr))
    {
      /* f is probably not open, so ignore errors */
      (void)close(f);
    }
  }
  errno = 0;
  return 0;
}

static int make_pidfile(const char *dirname)
{
  char *pidfilename = new char[ strlen(dirname)+10 ];
  strcpy(pidfilename,dirname);
  strcat(pidfilename, "/noad.pid");
  /*
  int pidf = open(pidfilename,O_RDONLY);
  if( pidf > 0 )
  {
    close(pidf);
    fprintf(stderr,"There is already running an instance of noad for this recording!\n");
    fprintf(stderr,"If you feel that this is not true, you have to delete the file\n");
    fprintf(stderr,"%s manually\n",pidfilename);
    return 1;
  }
  */
  int pidf = open(pidfilename, O_RDWR|O_CREAT|O_EXCL, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
  if( pidf < 0 )
  {
    if( errno == EEXIST )
    {
      fprintf(stderr,"There is already running an instance of noad for this recording!\n");
      fprintf(stderr,"If you feel that this is not true, you have to delete the file\n");
      fprintf(stderr,"%s manually\n",pidfilename);
      return -1;
    }
    else
    {
      perror("can't create pidfile!");
      return -2;
    }
  }
  sprintf(pidfilename,"%d",getpid());
  write(pidf,pidfilename,strlen(pidfilename));
  close(pidf);
  return 0;
}

static int rm_pidfile(const char *dirname)
{
  char *pidfilename = new char[ strlen(dirname)+10 ];
  strcpy(pidfilename,dirname);
  strcat(pidfilename, "/noad.pid");
  unlink(pidfilename);
  return 0;
}

#ifndef HAVE_STRSIGNAL
static char signumbuf[10];
char *strsignal(int sig)
{
  sprintf(signumbuf,"%d",sig);
  return signumbuf;
}
#endif
static void signal_handler(int sig)
{
  if (!bFork)
    fprintf(stderr, "Aborted by signal %s...\n", strsignal(sig));
  rm_pidfile(recDir);
  syslog(LOG_INFO, "noad aborted by signal %s",strsignal(sig));
  exit(EXIT_FAILURE);
}

int main(int argc, char *argv[], char *envp[])
{
  //
  for(int i = 0; i < argc; i++)
    syslog(LOG_INFO, "noad arg[%d]: %s", i, argv[i]);
  syslog(LOG_INFO, "noad args done");

  //syslog(LOG_INFO, "noad pid: %d", getpid());

  int c;
  bool bImmediateCall = false;
  bool bAfter = false;
  bool bBefore = false;
  bool bEdited = false;
  bool bNice = false;
  int verbosity = 1;
  int niceLevel = 19; // changed to 19 (0.4.2)
  const char *logoCacheDir = NULL;
  const char *statFile = NULL;

  while (1)
  {
    int option_index = 0;
    static struct option long_options[] =
    {
      {"statisticfile", 1, 0, 's'},
      {"logocachedir", 1, 0, 'l'},
      {"verbose", 0, 0, 'v'},
      {"background", 0, 0, 'b'},
      {"priority",1,0,'p'},
      {"comments", 0, 0, 'c'},
      {"jumplogo",0,0,'j'},
      {"overlap",0,0,'o' },
      {"ac3",0,0,'a' },
      {"OSD",0,0,'O' },
      {"savelogo", 0, 0, 'S'},
      {"backupmarks", 0, 0, 'B'},
      {"scenechangedetection", 0, 0, 'C'},
      {"version", 0, 0, 'V'},
      {"nelonen",0,0,'n'},
      {0, 0, 0, 0}
    };

    c = getopt_long  (argc, argv, "s:l:vbp:cjoaOSBCV",
      long_options, &option_index);
    if (c == -1)
    break;

    switch (c)
    {
      case 's':
        statFile = optarg;
      break;

      case 'l':
        logoCacheDir = optarg;
      break;

      case 'v':
        verbosity++;
      break;

      case 'b':
        bFork = true;
      break;

      case 'p':
        if (isnumber(optarg) || *optarg=='-')
          niceLevel = atoi(optarg);
        else
        {
          fprintf(stderr, "noad: invalid priority level: %s\n", optarg);
          return 2;
        }
        bNice = true;
      break;

      case 'c':
        markComments = true;
      break;

      case 'j':
        logoJumpDetection = true;
      break;

      case 'o':
        detectOverlaps = true;
      break;

      case 'a':
        useAudioDetection = true;
      break;

      case 'O':
        osdMsg = 1;
      break;

      case 'S':
        saveLogo = true;
      break;

      case 'B':
        backupMarks = true;
      break;

      case 'C':
        useSceneChangeDetection = true;
      break;

      case 'V':
        printf("noad %s - no advertising\n",getVersion());
        return 0;

      case 'n':
        isNelonen = true;
      break;

      case '?':
        printf("unknow option ?\n");
      break;

      default:
        printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }

  if (optind < argc)
  {
    while (optind < argc)
    {
      if(strcmp(argv[optind], "after" ) == 0 )
      {
        bAfter = bFork = bNice = true;
      }
      else if(strcmp(argv[optind], "before" ) == 0 )
      {
        bBefore = true;
      }
      else if(strcmp(argv[optind], "edited" ) == 0 )
      {
        bEdited = true;
      }
      else if(strcmp(argv[optind], "nice" ) == 0 )
        bNice = true;
      else if(strcmp(argv[optind], "-" ) == 0 )
        bImmediateCall = true;
      else
      {
        if( strstr(argv[optind],".rec") != NULL )
          recDir = argv[optind];
      }
      optind++;
    }
  }

  // we can run, if one of bImmediateCall, bAfter, bBefore or bNice is true
  // and recDir is given
  if( (bImmediateCall || bAfter || bBefore || bEdited || bNice) && recDir )
  {
    // do nothing if called from vdr before the recording has startet
    // or after the video is cutted
    if( bBefore || bEdited )
      return 0;

    // if bFork is given go in background
    if( bFork )
    {
      (void)umask((mode_t)0077);
      close_files();
      pid_t pid = fork();
      if (pid < 0)
      {
        fprintf(stderr, "%m\n");
        esyslog(LOG_ERR, "fork ERROR: %m");
        return 2;
      }
      if (pid != 0)
        return 0; // initial program immediately returns
    }
    if( make_pidfile(recDir) < 0 )
      exit(-1);
    if( bFork )
    {
      //syslog(LOG_INFO, "noad (forked) pid: %d", getpid());
      chdir("/");
      if (setsid() == (pid_t)(-1))
      {
          perror("setsid");
          exit(EXIT_FAILURE);
      }
      if (signal(SIGHUP, SIG_IGN) == SIG_ERR)
      {
          perror("signal(SIGHUP, SIG_IGN)");
          errno = 0;
      }
      int f;
      
      f = open("/dev/null", O_RDONLY);
      if (f == -1)
      {
          perror("/dev/null");
          errno = 0;
      }
      else
      {
          if (dup2(f, fileno(stdin)) == -1)
          {
              perror("dup2");
              errno = 0;
          }
          (void)close(f);
      }
      
      f = open("/dev/null", O_WRONLY);
      if (f == -1)
      {
          perror("/dev/null");
          errno = 0;
      }
      else
      {
          if (dup2(f, fileno(stdout)) == -1)
          {
              perror("dup2");
              errno = 0;
          }
          if (dup2(f, fileno(stderr)) == -1)
          {
              perror("dup2");
              errno = 0;
          }
          (void)close(f);
      }
    }


    int MaxPossibleFileDescriptors = getdtablesize();
    for (int i = STDERR_FILENO + 1; i < MaxPossibleFileDescriptors; i++)
        close(i); //close all dup'ed filedescriptors


    // should we renice ?
    if( bNice )
    {
      int niceErr = nice(niceLevel);
      if( errno == EPERM || errno == EACCES )
      {
        fprintf(stderr, "nice %d: no super-user rights\n",niceLevel);
        esyslog(LOG_ERR, "ERROR: nice %d: no super-user rights",niceLevel);
      }
      else if( niceErr != niceLevel )
      {
        fprintf(stderr, "%d %d %m\n",niceErr,errno);
        esyslog(LOG_ERR, "nice ERROR(%d,%d): %m",niceLevel,niceErr);
      }
    }

    // set the log-Level
    SysLogLevel = verbosity;

    // catch some signals
    signal(SIGINT, signal_handler);
    signal(SIGTERM, signal_handler);
    signal(SIGABRT, signal_handler);
    signal(SIGSEGV, signal_handler);
    signal(SIGKILL, signal_handler);
    
    // now do the work..,
    doNoad(false, recDir);

    // write statistic
    if( statFile != NULL )
       writeStatistic(statFile,recDir);

    rm_pidfile(recDir);
    return 0;
  }

  // nothing done, give the user some help
  printf("Usage: noad [options] cmd <record>\n"
         "options:\n"
         "-s <filename>,  --statisticfile=<file>\n"
         "                  filename where some statistic datas are stored\n"
         //"-l <directory>, --logocachedir=<directory>\n"
         "-v,             --verbose\n"
         "                  increments loglevel by one, can be given multiple\n"
         "-b,             --background\n"
         "                  noad runs as a background-process\n"
         "-p,             --priority\n"
         "                  priority-level of noad when running in background [19...-19] default 19\n"
         "-c,             --comments\n"
         "                  add comments to the marks\n"
         "-o,             --overlap\n"
         "                  detect overlaps\n"
         "-a,             --ac3\n"
         "                  use ac3-detection\n"
         "-j,             --jumplogo\n"
         "                  detect jumping logos\n"
         "-O,             --OSD\n"
         "                  noad sends an OSD-Message to localhost:2001 for start and end\n"
         "-S              --savelogo\n"
         "                  save the detected logo\n"
         "-B              --backupmarks\n"
         "                  move the marks.vdr to marks0.vdr\n"
         "-n              --nelonen\n"
         "                  special behavior for finish stations\n"
         //"-C              --scenechangedetection\n"
         //"                  use scene-change-detection\n"
         "-V              --version\n"
         "                  print version-info and exits\n"
         "\ncmd: one of\n"
         "edited                       from vdr if used in the -r option of vdr\n"
         "                             noad exits immediately if called with \"edited\"\n"
         "before                       from vdr if used in the -r option of vdr\n"
         "                             noad exits immediately if called with \"before\"\n"
         "after                        from vdr if used in the -r option of vdr\n"
         "-                            dummy-parameter if called directly\n"
         "nice                         runs noad with nice(19)\n"
         "\n<record>                     is the name of the directory where the recording\n"
         "                             is stored\n\n"
         );

}


